/*
 *  shellink.c
 *
 *  Create and edit Windows shortcut files (.LNK) under '95 and NT.
 *
 *  Compile by MS Visual C++:
 *    cl shellink.c /MD
 *
 *  Compile by Borland C++:
 *    bcc32 shellink.c ole2w32.lib
 *
 *  Written by Evseev (evseev@csa.ru) at Mar 2000
 */

#define R_OK  0

#ifdef _MSC_VER
#  pragma comment(lib,"ole32.lib")    /* CoXxx() */
#  pragma comment(lib,"uuid.lib")     /* IID_IPersistFile */
#  pragma comment(lib,"shell32.lib")  /* CLSID_ShellLink, IID_ShellLink */
#  pragma comment(lib,"user32.lib")   /* MessageBox() */
#  pragma warning(disable:4229)       /* allow __cdecl in data declarations */
#endif

#include <windows.h>
#include <shlobj.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>

#include "winentry.c"

typedef struct {
    IShellLink   *as_link;
    IPersistFile *as_file;
    char  linkname[MAX_PATH];

    char  arguments[MAX_PATH];
    char  description[MAX_PATH];
    WORD  hotkey;
    char  iconfile[MAX_PATH];
    int   icon_ndx;

    char  filename[MAX_PATH];
    WIN32_FIND_DATA finddata;
    DWORD findmode;   /* 0, SLGP_SHORTPATH or SLGP_UNCPRIORITY */
    int   is_relative;
 
    int   showcmd;
    char  directory[MAX_PATH];
} Shell_link;


HRESULT init_shell_link (Shell_link *data)
{
    HRESULT hr;
    if (!data)
        return E_INVALIDARG;
    memset (data, 0, sizeof(Shell_link));

    hr = CoCreateInstance( &CLSID_ShellLink,
                           NULL,
                           CLSCTX_INPROC_SERVER,
                           &IID_IShellLink,
                           (void**)&data->as_link );
    if (SUCCEEDED(hr)) {
        hr = data->as_link->lpVtbl->QueryInterface(
             data->as_link, &IID_IPersistFile, (void**)&data->as_file);
    }
    return hr;
}


HRESULT done_shell_link (Shell_link *data)
{
    if (!data)
        return E_INVALIDARG;
    if(data->as_file) if(!data->as_file->lpVtbl->Release(data->as_file)) data->as_file = NULL;
    if(data->as_link) if(!data->as_link->lpVtbl->Release(data->as_link)) data->as_link = NULL;
    return S_OK;
}


HRESULT load_shell_link (Shell_link *data)
{
    HRESULT hr;
    WORD ole_filename[MAX_PATH];
    if (!data)
        return E_INVALIDARG;

    // Ensure that the string is ANSI.
    MultiByteToWideChar(CP_ACP, 0, data->linkname, -1, ole_filename, MAX_PATH);
    hr = data->as_file->lpVtbl->Load(data->as_file, ole_filename, STGM_READ);
    if (FAILED(hr))
        return hr;
                
    data->as_link->lpVtbl->GetArguments       (data->as_link, data->arguments,   MAX_PATH);
    data->as_link->lpVtbl->GetDescription     (data->as_link, data->description, MAX_PATH);
    data->as_link->lpVtbl->GetHotkey          (data->as_link, &data->hotkey);
    data->as_link->lpVtbl->GetIconLocation    (data->as_link, data->iconfile,    MAX_PATH, &data->icon_ndx );
    data->as_link->lpVtbl->GetPath            (data->as_link, data->filename,    MAX_PATH, &data->finddata, data->findmode );
    data->as_link->lpVtbl->GetShowCmd         (data->as_link, &data->showcmd);
    data->as_link->lpVtbl->GetWorkingDirectory(data->as_link, data->directory,   MAX_PATH);

    return hr;
}


HRESULT save_shell_link (Shell_link *data)
{
    HRESULT hr;
    WORD ole_filename[MAX_PATH];

    if (!data)
        return E_INVALIDARG;

    data->as_link->lpVtbl->SetDescription     (data->as_link, data->description);
    data->as_link->lpVtbl->SetHotkey          (data->as_link, data->hotkey);
    data->as_link->lpVtbl->SetIconLocation    (data->as_link, data->iconfile, data->icon_ndx);
    data->as_link->lpVtbl->SetWorkingDirectory(data->as_link, data->directory);
    data->as_link->lpVtbl->SetArguments       (data->as_link, data->arguments);
  if (data->is_relative) {
    data->as_link->lpVtbl->SetRelativePath    (data->as_link, data->filename, 0);
  } else {
    data->as_link->lpVtbl->SetPath            (data->as_link, data->filename);
  }
    data->as_link->lpVtbl->SetShowCmd         (data->as_link, data->showcmd);

    // Ensure that the string is ANSI.
    MultiByteToWideChar(CP_ACP, 0, data->linkname, -1, ole_filename, MAX_PATH);
    hr = data->as_file->lpVtbl->Save(data->as_file, ole_filename, TRUE);
    if (FAILED(hr))
        return hr;

    return hr;
}


HRESULT dump_shell_link (Shell_link *data)
{
    if (!data)
        return E_INVALIDARG;
    echo(NULL,

        "target file  : \"%s\"\n"
        "description  : \"%s\"\n"
        "command line : \"%s\"\n"
        "directory    : \"%s\"\n"
        "hotkey       : %d (hex %x)\n"
        "icon         : \"%s\",%d\n"
        "showcmd      : %d\n",

        data->filename,
        data->description,
        data->arguments,
        data->directory,
        data->hotkey, data->hotkey,
        data->iconfile,
        data->icon_ndx,
        data->showcmd
    );
    return S_OK;
}


void stop(char *message, DWORD exitcode)
{
    echo(NULL, (FAILED(exitcode) ? "%s, error = %x"  :
               (    exitcode     ? "%s, error = %ld" : "%s")),
        message, exitcode);
    exit(exitcode);
}


void usage (char *error_message)
{
    char buf[1024];
    int pos = 0;
    if (error_message)
        pos = sprintf(buf, "ERROR: %s\n", error_message);
    sprintf( buf+pos,
        "Usage:\n"
        "   shellink {dump|new|replace|change} file.lnk [options...]\n"
        "Options:\n"
        "   -t filepath     full (path and name) filename of target object\n"
      //"   -f filename     relative filename of target object\n"
        "   -a arguments    command line\n"
        "   -e description  \n"
        "   -k hotkey       integer 16-bit value\n"
        "   -i iconfile     \n"
        "   -x icon_ndx     integer index of icon in iconfile\n"
        "   -d workdir      \n"
        "   -w showcmd      integer value (Norm=1, Min=7, Max=3)\n"
        "   -r              Resolve broken shortcut\n"
        "Written by ilya_evseev@chat.ru at Mar 2000");
    echo(NULL, buf);
    exit(1);
}


Shell_link shell_link;

void exit_routine(void) {
    done_shell_link(&shell_link);
    CoUninitialize();
}


int main_body (int argc, char *argv[])
{
    HRESULT hr;
    LPTSTR unused;
    register char *p;
    int i;
    enum { DUMP, CREATE, REPLACE, CHANGE }
        command;
    int resolve_broken = 0;

    hr = CoInitialize(NULL);
    if (FAILED(hr))
        stop("CoInitialize failed", hr);
    init_shell_link (&shell_link);
    atexit (exit_routine);

    if (argc < 2)
        usage(NULL);
    if (argc < 3)
        usage("missing link name");

    /*---------------------------------------------------------
     *  parse command (first parameter)
     */
    p = argv[1];
    if (!strcmp(p,"help") || !strcmp(p,"-h") || !strcmp(p,"-?") || !strcmp(p,"/?")) {
        usage(NULL);
    } else
    if (!strcmp(p,"dump") || !strcmp(p,"list") || !strcmp(p,"l")) {
        command = DUMP;
    } else
    if (!strcmp(p,"create") || !strcmp(p,"new") || !strcmp(p,"n")) {
        command = CREATE;
    } else
    if (!strcmp(p,"force") || !strcmp(p,"f") || !strcmp(p,"replace") || !strcmp(p,"r")) {
        command = REPLACE;
    } else
    if (!strcmp(p,"change") || !strcmp(p,"c")) {
        command = CHANGE;
    } else {
        usage("Unknown command");
    }

    if (!GetFullPathName (argv[2], MAX_PATH, shell_link.linkname, &unused))
        stop("cannot get full path name of shell link", GetLastError());

    /*---------------------------------------------------------
     *  load existing shortcut
     */
    switch (command) {
        case REPLACE:
            break;
        case CREATE:
            if (access(shell_link.linkname, R_OK) == 0)
                stop("Cannot overwrite existing link file", 0);
            break;
        case CHANGE:
        case DUMP:
            hr = load_shell_link(&shell_link);
            if (FAILED(hr) && (command == DUMP))
                stop("Loading link file failed", hr);
            break;
    }

    switch (command) {
    case DUMP:
        dump_shell_link(&shell_link);
        break;
    case CREATE:
    case REPLACE:    /* link contents must be saved */
    case CHANGE:
        /*-----------------------------------------------------
         *  parse options
         */
        for (i = 3; i < argc; i++) {
            p = argv[i];
            if (!strcmp(p,"-t"))
         // || (!strcmp(p,"-f") && ((shell_link.is_relative = 1) != 0)))
            {
              //if (*shell_link.filename)    usage("target name given twice");
                if (++i >= argc)             usage("missing target name");
                if (!GetFullPathName(argv[i], MAX_PATH, shell_link.filename, &unused))
                    stop("cannot get full path name of target", GetLastError());
            } else
            if (!strcmp(p,"-a")) {
              //if (*shell_link.arguments)   usage("arguments given twice");
                if (++i >= argc)             usage("missing arguments");
                strcpy(shell_link.arguments, argv[i]);
            } else
            if (!strcmp(p,"-e")) {
              //if (*shell_link.description) usage("description given twice");
                if (++i >= argc)             usage("missing description");
                strcpy(shell_link.description, argv[i]);
            } else
            if (!strcmp(p,"-k")) {
              //if (shell_link.hotkey)       usage("hotkey given twice");
                if (++i >= argc)             usage("missing hotkey");
                if ((shell_link.hotkey = atoi(argv[i])) == 0)
                    usage("invalid hotkey");
            } else
            if (!strcmp(p,"-i")) {
              //if (*shell_link.iconfile)    usage("icon file given twice");
                if (++i >= argc)             usage("missing icon file");
                if (!GetFullPathName (argv[i], MAX_PATH, shell_link.iconfile, &unused))
                    stop("cannot get full path name of iconfile", GetLastError());
            } else
            if (!strcmp(p,"-x")) {
              //if (shell_link.icon_ndx)     usage("icon index given twice");
                if (++i >= argc)             usage("missing icon index");
                shell_link.icon_ndx = atoi(argv[i]);
            } else
            if (!strcmp(p,"-d")) {
              //if (*shell_link.directory)   usage("directory given twice");
                if (++i >= argc)             usage("missing directory");
                if (!GetFullPathName (argv[i], MAX_PATH, shell_link.directory, &unused))
                    stop("cannot get full path name of directory", GetLastError());
            } else
            if (!strcmp(p,"-w")) {
              //if (shell_link.showcmd)      usage("showcmd given twice");
                if (++i >= argc)             usage("missing showcmd");
                shell_link.showcmd = atoi(argv[i]);
            } else
            if (!strcmp(p,"-r")) {
                resolve_broken = 1;
            }
        }

        if (!*shell_link.filename)
            stop("Target filename required, but is empty now", 1);

        if (resolve_broken) {
            hr = shell_link.as_link->lpVtbl->Resolve(
                 shell_link.as_link, NULL, SLR_ANY_MATCH);
            if (FAILED(hr))
                stop("Resolving failed", hr);
        }

        hr = save_shell_link (&shell_link);
        if (FAILED(hr))
            stop("Cannot save link file", hr);
        break; /* switch */
    } /* if link contents must be saved */

    return 0;
}

/* EOF */
