/*
 *  winentry.c
 *
 *  Dual startup support for Windows apps: GUI and console
 *
 *  Checks __CONSOLE__ macro at compile time
 *  and generates appropriate entry point
 *
 *  In both cases calls external routine:
 *     extern int main_body(int argc, char *argv[])
 *  that must be implemented by user
 *
 *  How to build demo test by Borland C++ 5:
 *     bcc32 -DTEST_WINENTRY -tWC winentry.c    ...generates console app
 *     bcc32 -DTEST_WINENTRY -tW  winentry.c    ...generates GUI app
 *  ..by Visual C:
 *     cl /MD /DTEST_WINENTRY /D__CONSOLE__ winentry.c    ...console app
 *     cl /MD /DTEST_WINENTRY winentry.c                  ...GUI app
 *
 *  Written at Feb 2000 by Evseev
 */

extern int main_body(int argc, char *argv[]);

#include <stdio.h>   /* sprintf - used by ABORT_IF */
#include <stdarg.h>

#ifndef __CONSOLE__

#   include <windows.h>

    int parse_command_line(char **dest_argv, int argv_size, char *src_cmdline)
    {
        int n = 0;
        int wait_quote = 0;
        char *p = src_cmdline;
        while (n < argv_size) {
            while (*p == ' ' || *p == '\t')
                p++;
            if (!*p)
                break;
            dest_argv[n++] = p;
            while (*p) {
                if (*p == '"') {
                    wait_quote = !wait_quote;
                    strcpy(p, p+1);
                    continue;
                }
                if (!wait_quote && (*p == ' ' || *p == '\t'))
                    break;
                p++;
            }
            if (!*p)
                break;
            *p++ = 0;
        }
        return n;
    }

#   ifdef __BORLANDC__
#     pragma argsused
#   endif

    int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
        LPSTR lpszCmdLine, int nCmdShow)	
    {
        char *argv[32];
        char module_filename[1024];
        GetModuleFileName(NULL, argv[0] = module_filename, 1024);
        return main_body(1+parse_command_line(argv+1, 31, lpszCmdLine), argv);
    }

#else   /* !ifndef __CONSOLE__ */

    int main(int argc, char *argv[])
    {
        return main_body(argc, argv);
    }

#endif   /* !ifndef __CONSOLE__ */


#ifdef __CONSOLE__

int echo(char *title, char *format_message, ...)
{
    char s[1024];
    char *p = s;
    va_list argptr;

    if (title)
        p += sprintf(s, "%s: ", title);

    va_start(argptr, format_message);
    vsprintf(p, format_message, argptr);
    va_end(argptr);
    puts(s);
    return 0;
}

#else   /* !ifdef __CONSOLE__ */

int echo(char *title, char *format_message, ...)
{
    char s[1024];
    va_list argptr;

    if (!title)
        title = "Notification";

    va_start(argptr, format_message);
    vsprintf(s, format_message, argptr);
    va_end(argptr);

    MessageBox(NULL, s, title, MB_ICONINFORMATION | MB_OK);
    return 0;
}

#endif   /* !ifdef __CONSOLE__ */

static char ABORT_IF_fmt[] = "%s: error %ld";
static char ABORT_IF_err[] = "ERROR";

#define ABORT_IF(Condition, Hint)  \
        do {  \
            if (Condition) {  \
                echo(ABORT_IF_err, ABORT_IF_fmt, Hint, GetLastError());  \
                exit(1);  \
            }  \
        } while(0)


#ifdef TEST_WINENTRY

#include <stdio.h>   /* sprintf, printf */
#include <string.h>  /* strlen */

int main_body(int argc, char *argv[])
{
    char s[1024];
    int i;
    *s = 0;
    for (i = 0; i < argc; i++) {
        sprintf(s + strlen(s), "\n%d = %s", i, argv[i]);
    }
    echo("Command line args", s);
/*
#ifdef __CONSOLE__
    printf(s);
#else
    MessageBox(NULL, s, "Command line args:", MB_OK);
#endif
*/
    return 0;
}

#endif  /* TEST_WINENTRY */

/* EOF */
